=== modified file 'bzrlib/smart/__init__.py'
--- bzrlib/smart/__init__.py	2007-04-26 04:32:44 +0000
+++ bzrlib/smart/__init__.py	2007-09-03 04:43:27 +0000
@@ -20,160 +20,17 @@
 rather than being a single large module.  Refer to the individual module
 docstrings for details.
 
-Overview
-========
-
-The smart protocol provides a way to send a requests and corresponding
-responses to communicate with a remote bzr process.
-
-Layering
-========
-
-Medium
-------
-
-At the bottom level there is either a socket, pipes, or an HTTP
-request/response.  We call this layer the *medium*.  It is responsible for
-carrying bytes between a client and server.  For sockets, we have the
-idea that you have multiple requests and get a read error because the other side
-did shutdown.  For pipes we have read pipe which will have a zero read which
-marks end-of-file.  For HTTP server environment there is no end-of-stream
-because each request coming into the server is independent.
-
-So we need a wrapper around pipes and sockets to seperate out requests from
-substrate and this will give us a single model which is consistent for HTTP,
-sockets and pipes.
-
-Protocol
---------
-
-On top of the medium is the *protocol*.  This is the layer that deserialises
-bytes into the structured data that requests and responses consist of.
-
-Version one of the protocol (for requests and responses) is described by::
-
-  REQUEST := MESSAGE_V1
-  RESPONSE := MESSAGE_V1
-  MESSAGE_V1 := ARGS BODY
-
-  ARGS := ARG [MORE_ARGS] NEWLINE
-  MORE_ARGS := SEP ARG [MORE_ARGS]
-  SEP := 0x01
-
-  BODY := LENGTH NEWLINE BODY_BYTES TRAILER
-  LENGTH := decimal integer
-  TRAILER := "done" NEWLINE
-
-That is, a tuple of arguments separated by Ctrl-A and terminated with a newline,
-followed by length prefixed body with a constant trailer.  Note that although
-arguments are not 8-bit safe (they cannot include 0x01 or 0x0a bytes without
-breaking the protocol encoding), the body is.
-
-Version two of the request protocol is::
-
-  REQUEST_V2 := "bzr request 2" NEWLINE MESSAGE_V1
-
-Version two of the response protocol is::
-
-  RESPONSE_V2 := "bzr request 2" NEWLINE MESSAGE_V1
-
-Future versions should follow this structure, like version two does::
-
-  FUTURE_MESSAGE := VERSION_STRING NEWLINE REST_OF_MESSAGE
-
-This is that clients and servers can read bytes up to the first newline byte to
-determine what version a message is.
-
-For compatibility will all versions (past and future) of bzr clients, servers
-that receive a request in an unknown protocol version should respond with a
-single-line error terminated with 0x0a (NEWLINE), rather than structured
-response prefixed with a version string.
-
-Request/Response processing
----------------------------
-
-On top of the protocol is the logic for processing requests (on the server) or
-responses (on the client).
-
-Server-side
------------
-
-Sketch::
-
- MEDIUM  (factory for protocol, reads bytes & pushes to protocol,
-          uses protocol to detect end-of-request, sends written
-          bytes to client) e.g. socket, pipe, HTTP request handler.
-  ^
-  | bytes.
-  v
-
- PROTOCOL(serialization, deserialization)  accepts bytes for one
-          request, decodes according to internal state, pushes
-          structured data to handler.  accepts structured data from
-          handler and encodes and writes to the medium.  factory for
-          handler.
-  ^
-  | structured data
-  v
-
- HANDLER  (domain logic) accepts structured data, operates state
-          machine until the request can be satisfied,
-          sends structured data to the protocol.
-
-Request handlers are registered in `bzrlib.smart.request`.
-
-
-Client-side
------------
-
-Sketch::
-
- CLIENT   domain logic, accepts domain requests, generated structured
-          data, reads structured data from responses and turns into
-          domain data.  Sends structured data to the protocol.
-          Operates state machines until the request can be delivered
-          (e.g. reading from a bundle generated in bzrlib to deliver a
-          complete request).
-
-          Possibly this should just be RemoteBzrDir, RemoteTransport,
-          ...
-  ^
-  | structured data
-  v
-
-PROTOCOL  (serialization, deserialization)  accepts structured data for one
-          request, encodes and writes to the medium.  Reads bytes from the
-          medium, decodes and allows the client to read structured data.
-  ^
-  | bytes.
-  v
-
- MEDIUM  (accepts bytes from the protocol & delivers to the remote server.
-          Allows the potocol to read bytes e.g. socket, pipe, HTTP request.
-
-The domain logic is in `bzrlib.remote`: `RemoteBzrDir`, `RemoteBranch`, and so
-on.
+Server-side request handlers are registered in the `bzrlib.smart.request`
+module.
+
+The domain logic is in `bzrlib.remote`: `RemoteBzrDir`, `RemoteBranch`,
+and so on.
 
 There is also an plain file-level transport that calls remote methods to
 manipulate files on the server in `bzrlib.transport.remote`.
 
-Paths
-=====
-
-Paths are passed across the network.  The client needs to see a namespace that
-includes any repository that might need to be referenced, and the client needs
-to know about a root directory beyond which it cannot ascend.
-
-Servers run over ssh will typically want to be able to access any path the user
-can access.  Public servers on the other hand (which might be over http, ssh
-or tcp) will typically want to restrict access to only a particular directory
-and its children, so will want to do a software virtual root at that level.
-In other words they'll want to rewrite incoming paths to be under that level
-(and prevent escaping using ../ tricks.)
-
-URLs that include ~ should probably be passed across to the server verbatim
-and the server can expand them.  This will proably not be meaningful when
-limited to a directory?
+The protocol is described in doc/developers/network-protocol.txt.
+
 """
 
 # TODO: _translate_error should be on the client, not the transport because

=== modified file 'bzrlib/smart/protocol.py'
--- bzrlib/smart/protocol.py	2007-08-30 15:32:50 +0000
+++ bzrlib/smart/protocol.py	2007-09-03 08:28:51 +0000
@@ -18,6 +18,7 @@
 client and server.
 """
 
+import collections
 from cStringIO import StringIO
 import time
 
@@ -209,15 +210,27 @@
             bytes = self._encode_bulk_data(response.body)
             self._write_func(bytes)
         elif response.body_stream is not None:
-            for chunk in response.body_stream:
-                assert isinstance(chunk, str), 'body must be a str'
-                if chunk == '':
-                    # Skip empty chunks, as they would prematurely signal
-                    # end-of-stream, and they're redundant anyway.
-                    continue
-                bytes = "%x\n%s" % (len(chunk), chunk)
-                self._write_func(bytes)
-            self._write_func('0\n')
+            _send_stream(response.body_stream, self._write_func)
+
+
+def _send_stream(stream, write_func):
+    _send_chunks(stream, write_func)
+    write_func('END\n')
+
+
+def _send_chunks(stream, write_func):
+    for chunk in stream:
+        if isinstance(chunk, str):
+            bytes = "%x\n%s" % (len(chunk), chunk)
+            write_func(bytes)
+        elif isinstance(chunk, request.FailedSmartServerResponse):
+            write_func('ERR\n')
+            _send_chunks(chunk.args, write_func)
+            return
+        else:
+            raise BzrError(
+                'Chunks must be str or FailedSmartServerResponse, got %r'
+                % chunks)
 
 
 class _StatefulDecoder(object):
@@ -247,33 +260,26 @@
 class ChunkedBodyDecoder(_StatefulDecoder):
     """Decoder for chunked body data.
 
-    This is very similar the HTTP's chunked encoding::
-
-        BODY := CHUNKS TERMINATOR
-        CHUNKS := CHUNK [CHUNKS]
-        CHUNK := CHUNK_LENGTH CHUNK_CONTENT
-        CHUNK_LENGTH := HEX_DIGITS NEWLINE
-        CHUNK_CONTENT := bytes
-        TERMINATOR := '0' NEWLINE
-
-    That is, the body consists of a series of chunks.  Each chunk starts with a
-    length prefix in hexadecimal digits, followed by an ASCII newline byte.
-    The end of the body is signaled by a zero the zero-length chunk, i.e.
-    '0\\n'.
+    This is very similar the HTTP's chunked encoding.  See the description of
+    streamed body data in `doc/developers/network-protocol.txt` for details.
     """
 
     def __init__(self):
         _StatefulDecoder.__init__(self)
         self.state_accept = self._state_accept_expecting_length
         self._in_buffer = ''
-        self._content_bytes = ''
+        self.chunk_in_progress = None
+        self.chunks = collections.deque()
+        self.error = False
+        self.error_in_progress = None
     
     def next_read_size(self):
-        # Note: the shortest possible chunk is 2 bytes: '0\n'.
+        # Note: the shortest possible chunk is 2 bytes: '0\n', and the
+        # end-of-body marker is 4 bytes: 'END\n'.
         if self.state_accept == self._state_accept_reading_chunk:
             # We're expecting more chunk content.  So we're expecting at least
-            # the rest of this chunk plus another chunk header.
-            return self.bytes_left + 2
+            # the rest of this chunk plus an END chunk.
+            return self.bytes_left + 4
         elif self.state_accept == self._state_accept_expecting_length:
             if self._in_buffer == '':
                 # We're expecting a chunk length.  There's at least two bytes
@@ -288,40 +294,70 @@
         else:
             raise AssertionError("Impossible state: %r" % (self.state_accept,))
 
-    def read_pending_data(self):
-        bytes = self._content_bytes
-        self._content_bytes = ''
-        return bytes
+    def read_next_chunk(self):
+        try:
+            return self.chunks.popleft()
+        except IndexError:
+            return None
 
-    def _state_accept_expecting_length(self, bytes):
-        self._in_buffer += bytes
+    def _extract_line(self):
         pos = self._in_buffer.find('\n')
         if pos == -1:
             # We haven't read a complete length prefix yet, so there's nothing
             # to do.
-            return
-        self.bytes_left = int(self._in_buffer[:pos], 16)
-        # Trim the length and '\n' delimiter from the _in_buffer.
+            return None
+        line = self._in_buffer[:pos]
+        # Trim the prefix (including '\n' delimiter) from the _in_buffer.
         self._in_buffer = self._in_buffer[pos+1:]
-        if self.bytes_left == 0:
+        return line
+
+    def _finished(self):
+        self.unused_data = self._in_buffer
+        self._in_buffer = None
+        self.state_accept = self._state_accept_reading_unused
+        if self.error:
+            error_args = tuple(self.error_in_progress)
+            self.chunks.append(request.FailedSmartServerResponse(error_args))
+            self.error_in_progress = None
+        self.finished_reading = True
+
+    def _state_accept_expecting_length(self, bytes):
+        self._in_buffer += bytes
+        prefix = self._extract_line()
+        if prefix is None:
+            # We haven't read a complete length prefix yet, so there's nothing
+            # to do.
+            return
+        elif prefix == 'ERR':
+            self.error = True
+            self.error_in_progress = []
+            self._state_accept_expecting_length('')
+            return
+        elif prefix == 'END':
+            # We've read the end-of-body marker.
             # Any further bytes are unused data, including the bytes left in
             # the _in_buffer.
-            self.unused_data = self._in_buffer
-            self._in_buffer = None
-            self.state_accept = self._state_accept_reading_unused
-            self.finished_reading = True
+            self._finished()
             return
-        self.state_accept = self._state_accept_reading_chunk
+        else:
+            self.bytes_left = int(prefix, 16)
+            self.chunk_in_progress = ''
+            self.state_accept = self._state_accept_reading_chunk
 
     def _state_accept_reading_chunk(self, bytes):
         self._in_buffer += bytes
         in_buffer_len = len(self._in_buffer)
-        self._content_bytes += self._in_buffer[:self.bytes_left]
+        self.chunk_in_progress += self._in_buffer[:self.bytes_left]
         self._in_buffer = self._in_buffer[self.bytes_left:]
         self.bytes_left -= in_buffer_len
         if self.bytes_left <= 0:
             # Finished with chunk
             self.bytes_left = None
+            if self.error:
+                self.error_in_progress.append(self.chunk_in_progress)
+            else:
+                self.chunks.append(self.chunk_in_progress)
+            self.chunk_in_progress = None
             self.state_accept = self._state_accept_expecting_length
         
     def _state_accept_reading_unused(self, bytes):
@@ -560,7 +596,7 @@
         return SmartClientRequestProtocolOne.read_response_tuple(self, expect_body)
 
     def _write_protocol_version(self):
-        r"""Write any prefixes this protocol requires.
+        """Write any prefixes this protocol requires.
         
         Version two sends the value of REQUEST_VERSION_TWO.
         """
@@ -574,10 +610,9 @@
             bytes_wanted = _body_decoder.next_read_size()
             bytes = self._request.read_bytes(bytes_wanted)
             _body_decoder.accept_bytes(bytes)
-            body_bytes = _body_decoder.read_pending_data()
-            if body_bytes != '':
+            for body_bytes in iter(_body_decoder.read_next_chunk, None):
                 if 'hpss' in debug.debug_flags:
-                    mutter('              %d streamed body bytes read',
+                    mutter('              %d byte chunk read',
                            len(body_bytes))
                 yield body_bytes
         self._request.finished_reading()

=== modified file 'bzrlib/smart/repository.py'
--- bzrlib/smart/repository.py	2007-08-30 14:56:51 +0000
+++ bzrlib/smart/repository.py	2007-09-03 08:28:51 +0000
@@ -262,7 +262,6 @@
                 yield m[0]
                 m[0] = ''
         except errors.RevisionNotPresent, e:
-            yield 'ERROR'
-            #return FailedSmartServerResponse(('NoSuchRevision', e.revision_id))
+            yield FailedSmartServerResponse(('NoSuchRevision', e.revision_id))
         pack.end()
 

=== modified file 'bzrlib/smart/request.py'
--- bzrlib/smart/request.py	2007-08-30 14:24:58 +0000
+++ bzrlib/smart/request.py	2007-09-03 08:28:51 +0000
@@ -106,8 +106,8 @@
                 other.body_stream is self.body_stream)
 
     def __repr__(self):
-        return "<SmartServerResponse args=%r body=%r>" % (self.is_successful(),
-            self.args, self.body)
+        return ("<SmartServerResponse successful=%s args=%r body=%r>"
+                % (self.is_successful(), self.args, self.body))
 
 
 class FailedSmartServerResponse(SmartServerResponse):

=== modified file 'bzrlib/tests/test_smart.py'
--- bzrlib/tests/test_smart.py	2007-08-30 14:56:51 +0000
+++ bzrlib/tests/test_smart.py	2007-09-03 08:28:51 +0000
@@ -27,7 +27,11 @@
 import tarfile
 
 from bzrlib import bzrdir, errors, pack, smart, tests
-from bzrlib.smart.request import SmartServerResponse
+from bzrlib.smart.request import (
+    FailedSmartServerResponse,
+    SmartServerResponse,
+    SuccessfulSmartServerResponse,
+    )
 import bzrlib.smart.bzrdir
 import bzrlib.smart.branch
 import bzrlib.smart.repository
@@ -807,9 +811,16 @@
         repo = self.make_repository('.')
         rev_id1_utf8 = u'\xc8'.encode('utf-8')
         response = request.execute(backing.local_abspath(''), rev_id1_utf8)
+        # There's no error initially.
+        self.assertTrue(response.is_successful())
+        self.assertEqual(('ok',), response.args)
+        # We only get an error while streaming the body.
+        body = list(response.body_stream)
+        last_chunk = body[-1]
+        self.assertIsInstance(last_chunk, FailedSmartServerResponse)
         self.assertEqual(
-            SmartServerResponse(('NoSuchRevision', rev_id1_utf8)),
-            response)
+            last_chunk,
+            FailedSmartServerResponse(('NoSuchRevision', rev_id1_utf8)))
 
 
 class TestSmartServerIsReadonly(tests.TestCaseWithTransport):

=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2007-08-30 14:24:58 +0000
+++ bzrlib/tests/test_smart_transport.py	2007-09-03 08:28:51 +0000
@@ -1618,33 +1618,52 @@
                                       body_stream):
         """Assert that body_stream is serialised as expected_serialisation."""
         out_stream = StringIO()
-        smart_protocol = self.server_protocol_class(None, out_stream.write)
-        response = request.SuccessfulSmartServerResponse(
-            ('args',), body_stream=body_stream)
-        smart_protocol._send_response(response)
-        expected_prefix = protocol.RESPONSE_VERSION_TWO + 'success\n'
-        expected_args = 'args\n'
-        self.assertStartsWith(
-            out_stream.getvalue(), expected_prefix + expected_args)
-        body = out_stream.getvalue()[len(expected_prefix + expected_args):]
-        self.assertEqual(expected_serialisation, body)
+        protocol._send_stream(body_stream, out_stream.write)
+        self.assertEqual(expected_serialisation, out_stream.getvalue())
+
+    def assertBodyStreamRoundTrips(self, body_stream):
+        """Assert that body_stream is the same after being serialised and
+        deserialised.
+        """
+        out_stream = StringIO()
+        protocol._send_stream(body_stream, out_stream.write)
+        decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes(out_stream.getvalue())
+        decoded_stream = list(iter(decoder.read_next_chunk, None))
+        self.assertEqual(body_stream, decoded_stream)
 
     def test_body_stream_serialisation_empty(self):
         """A body_stream with no bytes can be serialised."""
-        self.assertBodyStreamSerialisation('0\n', [])
+        self.assertBodyStreamSerialisation('END\n', [])
+        self.assertBodyStreamRoundTrips([])
 
     def test_body_stream_serialisation(self):
+        stream = ['chunk one', 'chunk two', 'chunk three']
         self.assertBodyStreamSerialisation(
-            '9\nchunk one' + '9\nchunk two' + 'b\nchunk three' + '0\n',
-            ['chunk one', 'chunk two', 'chunk three'])
+            '9\nchunk one' + '9\nchunk two' + 'b\nchunk three' + 'END\n',
+            stream)
+        self.assertBodyStreamRoundTrips(stream)
 
     def test_body_stream_with_empty_element_serialisation(self):
-        """A body stream that includes '' does not prematurely end the stream.
+        """A body stream can include ''.
 
-        If the body stream yields an empty string, the protocol should not send
-        an empty chunk, because that is the signal for end-of-stream.
+        The empty string can be transmitted like any other string.
         """
-        self.assertBodyStreamSerialisation('5\nchunk' + '0\n', ['', 'chunk'])
+        stream = ['', 'chunk']
+        self.assertBodyStreamSerialisation(
+            '0\n' + '5\nchunk' + 'END\n', stream)
+        self.assertBodyStreamRoundTrips(stream)
+
+    def test_body_stream_error_serialistion(self):
+        stream = ['first chunk',
+                  request.FailedSmartServerResponse(
+                      ('FailureName', 'failure arg'))]
+        expected_bytes = (
+            'b\nfirst chunk' +
+            'ERR\n' + 'b\nFailureName' + 'b\nfailure arg' +
+            'END\n')
+        self.assertBodyStreamSerialisation(expected_bytes, stream)
+        self.assertBodyStreamRoundTrips(stream)
 
     def test_accept_bytes_of_bad_request_to_protocol(self):
         out_stream = StringIO()
@@ -1890,7 +1909,7 @@
 
     def test_streamed_body_bytes(self):
         two_body_chunks = "4\n1234" + "3\n567"
-        body_terminator = "0\n"
+        body_terminator = "END\n"
         server_bytes = (protocol.RESPONSE_VERSION_TWO +
                         "success\nok\n" + two_body_chunks + body_terminator)
         input = StringIO(server_bytes)
@@ -1903,6 +1922,28 @@
         stream = smart_protocol.read_streamed_body()
         self.assertEqual(['1234', '567'], list(stream))
 
+    def test_read_streamed_body_error(self):
+        """When a stream is interrupted by an error..."""
+        a_body_chunk = '4\naaaa'
+        err_signal = 'ERR\n'
+        err_chunks = 'a\nerror arg1' + '4\narg2'
+        finish = 'END\n'
+        body = a_body_chunk + err_signal + err_chunks + finish
+        server_bytes = (protocol.RESPONSE_VERSION_TWO +
+                        "success\nok\n" + body)
+        input = StringIO(server_bytes)
+        output = StringIO()
+        client_medium = medium.SmartSimplePipesClientMedium(input, output)
+        smart_request = client_medium.get_request()
+        smart_protocol = protocol.SmartClientRequestProtocolTwo(smart_request)
+        smart_protocol.call('foo')
+        smart_protocol.read_response_tuple(True)
+        expected_chunks = [
+            'aaaa',
+            request.FailedSmartServerResponse(('error arg1', 'arg2'))]
+        stream = smart_protocol.read_streamed_body()
+        self.assertEqual(expected_chunks, list(stream))
+
 
 class TestSmartClientUnicode(tests.TestCase):
     """_SmartClient tests for unicode arguments.
@@ -2015,15 +2056,15 @@
         decoder = protocol.ChunkedBodyDecoder()
         self.assertFalse(decoder.finished_reading)
         self.assertEqual(2, decoder.next_read_size())
-        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual(None, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
 
     def test_empty_content(self):
         """'0' + LF is the complete chunked encoding of a zero-length body."""
         decoder = protocol.ChunkedBodyDecoder()
-        decoder.accept_bytes('0\n')
+        decoder.accept_bytes('END\n')
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual(None, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
 
     def test_one_chunk(self):
@@ -2031,10 +2072,10 @@
         decoder = protocol.ChunkedBodyDecoder()
         chunk_length = 'f\n'
         chunk_content = '123456789abcdef'
-        finish = '0\n'
+        finish = 'END\n'
         decoder.accept_bytes(chunk_length + chunk_content + finish)
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual(chunk_content, decoder.read_pending_data())
+        self.assertEqual(chunk_content, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
         
     def test_incomplete_chunk(self):
@@ -2047,10 +2088,11 @@
         decoder.accept_bytes(chunk_length + three_bytes)
         self.assertFalse(decoder.finished_reading)
         self.assertEqual(
-            5 + 2, decoder.next_read_size(),
+            5 + 4, decoder.next_read_size(),
             "The next_read_size hint should be the number of missing bytes in "
-            "this chunk plus 2 (the shortest possible next chunk: '0\\n')")
-        self.assertEqual(three_bytes, decoder.read_pending_data())
+            "this chunk plus 4 (the length of the end-of-body marker: "
+            "'END\\n')")
+        self.assertEqual(None, decoder.read_next_chunk())
 
     def test_incomplete_length(self):
         """A chunk length hasn't been read until a newline byte has been read.
@@ -2063,31 +2105,33 @@
             "length yet.")
         decoder.accept_bytes('\n')
         self.assertEqual(
-            9 + 2, decoder.next_read_size(),
-            "The next_read_size hint should be the length of the chunk plus 2 "
-            "(the shortest possible next chunk: '0\\n')")
+            9 + 4, decoder.next_read_size(),
+            "The next_read_size hint should be the length of the chunk plus 4 "
+            "(the length of the end-of-body marker: 'END\\n')")
         self.assertFalse(decoder.finished_reading)
-        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual(None, decoder.read_next_chunk())
 
     def test_two_chunks(self):
         """Content from multiple chunks is concatenated."""
         decoder = protocol.ChunkedBodyDecoder()
         chunk_one = '3\naaa'
         chunk_two = '5\nbbbbb'
-        finish = '0\n'
+        finish = 'END\n'
         decoder.accept_bytes(chunk_one + chunk_two + finish)
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual('aaabbbbb', decoder.read_pending_data())
+        self.assertEqual('aaa', decoder.read_next_chunk())
+        self.assertEqual('bbbbb', decoder.read_next_chunk())
+        self.assertEqual(None, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
 
     def test_excess_bytes(self):
         """Bytes after the chunked body are reported as unused bytes."""
         decoder = protocol.ChunkedBodyDecoder()
-        chunked_body = "5\naaaaa0\n"
+        chunked_body = "5\naaaaaEND\n"
         excess_bytes = "excess bytes"
         decoder.accept_bytes(chunked_body + excess_bytes)
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual('aaaaa', decoder.read_pending_data())
+        self.assertEqual('aaaaa', decoder.read_next_chunk())
         self.assertEqual(excess_bytes, decoder.unused_data)
         self.assertEqual(
             1, decoder.next_read_size(),
@@ -2099,10 +2143,10 @@
         length = 0x123
         chunk_prefix = hex(length) + '\n'
         chunk_bytes = 'z' * length
-        finish = '0\n'
+        finish = 'END\n'
         decoder.accept_bytes(chunk_prefix + chunk_bytes + finish)
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual(chunk_bytes, decoder.read_pending_data())
+        self.assertEqual(chunk_bytes, decoder.read_next_chunk())
 
     def test_byte_at_a_time(self):
         """A complete body fed to the decoder one byte at a time should not
@@ -2115,11 +2159,11 @@
         decoder = protocol.ChunkedBodyDecoder()
         chunk_length = 'f\n'
         chunk_content = '123456789abcdef'
-        finish = '0\n'
+        finish = 'END\n'
         for byte in (chunk_length + chunk_content + finish):
             decoder.accept_bytes(byte)
         self.assertTrue(decoder.finished_reading)
-        self.assertEqual(chunk_content, decoder.read_pending_data())
+        self.assertEqual(chunk_content, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
 
     def test_read_pending_data_resets(self):
@@ -2127,12 +2171,25 @@
         decoder = protocol.ChunkedBodyDecoder()
         chunk_one = '3\naaa'
         chunk_two = '3\nbbb'
-        finish = '0\n'
+        finish = 'END\n'
         decoder.accept_bytes(chunk_one)
-        self.assertEqual('aaa', decoder.read_pending_data())
+        self.assertEqual('aaa', decoder.read_next_chunk())
         decoder.accept_bytes(chunk_two)
-        self.assertEqual('bbb', decoder.read_pending_data())
-        self.assertEqual('', decoder.read_pending_data())
+        self.assertEqual('bbb', decoder.read_next_chunk())
+        self.assertEqual(None, decoder.read_next_chunk())
+
+    def test_decode_error(self):
+        decoder = protocol.ChunkedBodyDecoder()
+        chunk_one = 'b\nfirst chunk'
+        error_signal = 'ERR\n'
+        error_chunks = '5\npart1' + '5\npart2'
+        finish = 'END\n'
+        decoder.accept_bytes(chunk_one + error_signal + error_chunks + finish)
+        self.assertTrue(decoder.finished_reading)
+        self.assertEqual('first chunk', decoder.read_next_chunk())
+        expected_failure = request.FailedSmartServerResponse(
+            ('part1', 'part2'))
+        self.assertEqual(expected_failure, decoder.read_next_chunk())
 
 
 class TestSuccessfulSmartServerResponse(tests.TestCase):
@@ -2147,6 +2204,8 @@
             ('foo', 'bar'), 'bytes')
         self.assertEqual(('foo', 'bar'), response.args)
         self.assertEqual('bytes', response.body)
+        # repr(response) doesn't trigger exceptions.
+        repr(response)
 
     def test_construct_with_body_stream(self):
         bytes_iterable = ['abc']
@@ -2176,6 +2235,8 @@
         response = request.FailedSmartServerResponse(('foo', 'bar'), 'bytes')
         self.assertEqual(('foo', 'bar'), response.args)
         self.assertEqual('bytes', response.body)
+        # repr(response) doesn't trigger exceptions.
+        repr(response)
 
     def test_is_successful(self):
         """is_successful should return False for FailedSmartServerResponse."""

=== modified file 'doc/developers/index.txt'
--- doc/developers/index.txt	2007-08-13 10:20:11 +0000
+++ doc/developers/index.txt	2007-09-03 05:11:38 +0000
@@ -31,5 +31,6 @@
 
 * `Repositories <repository.html>`_ |--| What repositories do and are used for.
 
+* `Network protocol <network-protocol.html>`_ |--| Custom network protocol.
 
 .. |--| unicode:: U+2014

=== added file 'doc/developers/network-protocol.txt'
--- doc/developers/network-protocol.txt	1970-01-01 00:00:00 +0000
+++ doc/developers/network-protocol.txt	2007-09-03 05:11:38 +0000
@@ -0,0 +1,223 @@
+================
+Network Protocol
+================
+
+:Date: 2007-09-03
+
+
+.. contents::
+
+
+Overview
+========
+
+The smart protocol provides a way to send a requests and corresponding
+responses to communicate with a remote bzr process.
+
+Layering
+========
+
+Medium
+------
+
+At the bottom level there is either a socket, pipes, or an HTTP
+request/response.  We call this layer the *medium*.  It is responsible for
+carrying bytes between a client and server.  For sockets, we have the idea
+that you have multiple requests and get a read error because the other
+side did shutdown.  For pipes we have read pipe which will have a zero
+read which marks end-of-file.  For HTTP server environment there is no
+end-of-stream because each request coming into the server is independent.
+
+So we need a wrapper around pipes and sockets to seperate out requests
+from substrate and this will give us a single model which is consistent
+for HTTP, sockets and pipes.
+
+Protocol
+--------
+
+On top of the medium is the *protocol*.  This is the layer that
+deserialises bytes into the structured data that requests and responses
+consist of.
+
+Request/Response processing
+---------------------------
+
+On top of the protocol is the logic for processing requests (on the
+server) or responses (on the client).
+
+Server-side
+-----------
+
+Sketch::
+
+ MEDIUM  (factory for protocol, reads bytes & pushes to protocol,
+          uses protocol to detect end-of-request, sends written
+          bytes to client) e.g. socket, pipe, HTTP request handler.
+  ^
+  | bytes.
+  v
+
+ PROTOCOL(serialization, deserialization)  accepts bytes for one
+          request, decodes according to internal state, pushes
+          structured data to handler.  accepts structured data from
+          handler and encodes and writes to the medium.  factory for
+          handler.
+  ^
+  | structured data
+  v
+
+ HANDLER  (domain logic) accepts structured data, operates state
+          machine until the request can be satisfied,
+          sends structured data to the protocol.
+
+Request handlers are registered in the `bzrlib.smart.request` module.
+
+
+Client-side
+-----------
+
+Sketch::
+
+ CLIENT   domain logic, accepts domain requests, generated structured
+          data, reads structured data from responses and turns into
+          domain data.  Sends structured data to the protocol.
+          Operates state machines until the request can be delivered
+          (e.g. reading from a bundle generated in bzrlib to deliver a
+          complete request).
+
+          Possibly this should just be RemoteBzrDir, RemoteTransport,
+          ...
+  ^
+  | structured data
+  v
+
+ PROTOCOL  (serialization, deserialization)  accepts structured data for one
+          request, encodes and writes to the medium.  Reads bytes from the
+          medium, decodes and allows the client to read structured data.
+  ^
+  | bytes.
+  v
+
+ MEDIUM  (accepts bytes from the protocol & delivers to the remote server.
+          Allows the potocol to read bytes e.g. socket, pipe, HTTP request.
+
+The domain logic is in `bzrlib.remote`: `RemoteBzrDir`, `RemoteBranch`,
+and so on.
+
+There is also an plain file-level transport that calls remote methods to
+manipulate files on the server in `bzrlib.transport.remote`.
+
+Protocol description
+====================
+
+Version one
+-----------
+
+Version one of the protocol (for requests and responses) is described by::
+
+  REQUEST := MESSAGE_V1
+  RESPONSE := MESSAGE_V1
+  MESSAGE_V1 := ARGS BODY
+
+  ARGS := ARG [MORE_ARGS] NEWLINE
+  MORE_ARGS := SEP ARG [MORE_ARGS]
+  SEP := 0x01
+
+  BODY := LENGTH NEWLINE BODY_BYTES TRAILER
+  LENGTH := decimal integer
+  TRAILER := "done" NEWLINE
+
+That is, a tuple of arguments separated by Ctrl-A and terminated with a
+newline, followed by length prefixed body with a constant trailer.  Note
+that although arguments are not 8-bit safe (they cannot include 0x01 or
+0x0a bytes without breaking the protocol encoding), the body is.
+
+Version two
+-----------
+
+Version two of the request protocol is::
+
+  REQUEST_V2 := "bzr request 2" NEWLINE MESSAGE_V2
+
+Version two of the response protocol is::
+
+  RESPONSE_V2 := "bzr request 2" NEWLINE MESSAGE_V2
+
+Future versions should follow this structure, like version two does::
+
+  FUTURE_MESSAGE := VERSION_STRING NEWLINE REST_OF_MESSAGE
+
+This is so that clients and servers can read bytes up to the first newline
+byte to determine what version a message is.
+
+For compatibility will all versions (past and future) of bzr clients,
+servers that receive a request in an unknown protocol version should
+respond with a single-line error terminated with 0x0a (NEWLINE), rather
+than structured response prefixed with a version string.
+
+Version two of the message protocol is::
+
+  MESSAGE_V2 := ARGS BODY
+  BODY_V2 := BODY | STREAMED_BODY
+
+That is, a version one length-prefixed body, or a version two streamed
+body.
+
+A streamed body looks a lot like HTTP's chunked encoding::
+ 
+  STREAMED_BODY := "chunked" NEWLINE CHUNKS TERMINATOR
+  CHUNKS := CHUNK [CHUNKS]
+  CHUNK := CHUNK_LENGTH CHUNK_CONTENT
+  CHUNK_LENGTH := HEX_DIGITS NEWLINE
+  CHUNK_CONTENT := bytes
+  
+  TERMINATOR := SUCCESS_TERMINATOR | ERROR_TERMINATOR
+  SUCCESS_TERMINATOR := 'END' NEWLINE
+  ERROR_TERMINATOR := 'ERR' NEWLINE CHUNKS SUCCESS_TERMINATOR
+
+That is, the body consists of a series of chunks.  Each chunk starts with
+a length prefix in hexadecimal digits, followed by an ASCII newline byte.
+The end of the body is signaled by 'END\\n', or by 'ERR\\n' followed by
+error args, one per chunk.  Note that this allows an 8-bit clean error
+response.
+
+A streamed body starts with the string "chunked" so that legacy clients
+and servers will not mistake the first chunk as the start of a version one
+body.
+
+Software versions
+-----------------
+
+Version 1 of the protocol was introduced in Bazaar 0.11.
+
+Version 2 was introduced in Bazaar 0.16.
+
+
+Paths
+=====
+
+Paths are passed across the network.  The client needs to see a namespace
+that includes any repository that might need to be referenced, and the
+client needs to know about a root directory beyond which it cannot ascend.
+
+Servers run over ssh will typically want to be able to access any path the
+user can access.  Public servers on the other hand (which might be over
+http, ssh or tcp) will typically want to restrict access to only a
+particular directory and its children, so will want to do a software
+virtual root at that level.  In other words they'll want to rewrite
+incoming paths to be under that level (and prevent escaping using ../
+tricks.)
+
+URLs that include ~ should probably be passed across to the server
+verbatim and the server can expand them.  This will proably not be
+meaningful when limited to a directory?
+
+
+Requests
+========
+
+The available request methods are registered in `bzrlib.smart.request`.
+
+..
+   vim: ft=rst tw=74 ai
+

